#!/usr/bin/env python
#
# Taglut : Topological And Geometrical Library - a Useful Toolkit
# Copyright (C) 2008 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
#                    CNRS / Univ. Blaise Pascal
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License 2
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


"""Usage: fastMarchingViewer [options] INPUT

Options:
    -n, --name         Object name
    -x, --x=X          3d window size (x)
    -y, --y=Y          3d window size (y)
    -p, --point=POINT  The starting point
    -r, --radius=R     Stop value
    -g, --opengl       Use opengl display rather than X11
    -b, --border       Show borders using red color
    --gouraud          Smooth the surface using gouraud mehtod
    -a, --autolight    Auto light of the surface

    -h, --help         Print this message and exit.

INPUT is a mesh file (available formats: VRML, PLY)"""

import taglut
import getopt
import sys

try:
    my_getopt = getopt.gnu_getopt
except AttributeError:
    my_getopt = getopt.getopt


def usage():
    print __doc__
    sys.exit(0)

def main():
    try:
        opts, args = my_getopt(sys.argv[1:], "n:hx:y:gbap:r:", ["name=", "help", "opengl",
                                                             "x=", "y=",
                                                             "gouraud", "autolight",
                                                             "point=", "radius=" ])
    except getopt.GetoptError, msg:
        # print help information and exit:
        print "Error: ", msg
        usage()
        sys.exit(2)

    opengl = False
    sizex = 300
    sizey = 300
    name = ""
    gouraud = False
    autolight = False
    point = 0
    radius = -1
    for o, a in opts:
        if o == "--gouraud":
            gouraud = True
        if o in ("-a", "--autolight"):
            autolight = True
        if o in ("-n", "--name"):
            name = a
        if o in ("-x", "--x"):
            sizex = int(a)
        if o in ("-y", "--y"):
            sizey = int(a)
        if o in ("-g", "--opengl"):
            opengl = True
        if o in ("-p", "--point"):
            point = int(a)
        if o in ("-r", "--radius"):
            radius = float(a)
        if o in ("-h", "--help"):
            usage()
            sys.exit()

    if sizex <= 0 or sizey <= 0:
        usage()
        print "Error: Output size not positive"
        sys.exit(2)

    try:
        inputFile = args[0]
    except IndexError:
        print "Error: Input file not defined"
        usage()
        sys.exit()

    try:
        print "Loading mesh"
        m = taglut.Mesh(inputFile, name)
        if point > m.getNbPoints() - 1:
            print "Point out of range (", m.getNbPoints(), ")"
            sys.exit(1)

        print "Computing distances"
        dTool = taglut.FastMarching()

        mMap = taglut.MeshMap(m)
        dTool.computeDistanceFromPoint(mMap, m, point, radius)

        v = 0
        for i in range(0, mMap.getNbValues() - 1):
            if mMap.getValue(i) == -1:
                print " V:"
                m.tracePoint(i)
                if m.isSaddlePoint(i):
                    print "   saddle point"
                for nb in m.point(i).getNeighbours():
                    print "  v(", nb, "): ", mMap.getValue(nb), ", d: ", m.point(nb).distance(m.point(i))
                    if m.isSaddlePoint(nb):
                        print "   saddle point"
                v += 1
        if v != 0:
            print "Warning: number of non-computed vertices:", v, "/", mMap.getNbValues()

        mMap.save("/tmp/map.txt")
        taglut.Display3D.displayMeshMapTriangles(sizex, sizey, "Display", m, mMap, 0.0, False, opengl, gouraud, autolight)

    except taglut.ExceptionFileNotFound, msg:
        print "Error: ", msg.getMessage()
        sys.exit()


if __name__ == "__main__":
    main()
