#!/usr/bin/env python
#
# Taglut : Topological And Geometrical Library - a Useful Toolkit
# Copyright (C) 2008 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
#                    CNRS / Univ. Blaise Pascal
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License 2
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.



"""Usage: meshCutUsingNLoops.py [options] INPUTS

Options:
    -a, --angles            Adjust using angles
    -c, --cylinders         Do not cut the cylinders after the n-loop computation
    -d, --display           Display the final mesh
    -i, --infos             Display information about final mesh
    -o, --output=FILE       Save the mesh using the file FILE
    -p, --parameterize=FILE Save the mesh and the parameterization using a Floater approach tile by tile
    -4, --quadrangulation   Describe the extrema tiles with quadrangles
    -x, --x=X               3d window size (x)
    -y, --y=Y               3d window size (y)
    -h, --help              Print this message and exit.

INPUTS is a mesh file, and a scalar function on the vertices"""


import taglut
import getopt
import sys

try:
    my_getopt = getopt.gnu_getopt
except AttributeError:
    my_getopt = getopt.getopt

def usage():
    print __doc__
    sys.exit(0)


def main():
    try:
        opts, args = my_getopt(sys.argv[1:], "hdio:x:y:acp:4", ["help", "infos", "display", "output=",
                                                               "x=", "y=", "angles", "cylinders",
                                                               "parameterize=", "quadrangulation"])
    except getopt.GetoptError, msg:
        # print help information and exit:
        print "Error: ", msg
        usage()

        sys.exit(2)

    infos = False
    display = False
    sizex = 300
    sizey = 300
    outputFile = ""
    outputFileParam = ""
    nLoops = False
    plnLoops = False
    angles = False
    cylinders = False
    quads = False
    for o, a in opts:
        if o in ("-i", "--infos"):
            infos = True
        if o in ("-d", "--display"):
            display = True
        if o in ("-x", "--x"):
            sizex = int(a)
        if o in ("-y", "--y"):
            sizey = int(a)
        if o in ("-o", "--output"):
            outputFile = a
        if o in ("-p", "--parameterize"):
            outputFileParam = a
        if o in ("-c", "--cylinders"):
            cylinders = True
        if o in ("-a", "--angles"):
            angles = True
        if o in ("-4", "--quadrangulation"):
            quads = True
        if o in ("-h", "--help"):
            usage()
            sys.exit()

    if not infos and not display and outputFile == "" and outputFileParam == "":
        print "No output defined. Abort"
        sys.exit()

    try:
        inputFile = args[0]
    except IndexError:
        print "Error: Mesh file not defined"
        usage()
        sys.exit()

    try:
        functionFile = args[1]
    except IndexError:
        print "Error: Function file not defined"
        usage()
        sys.exit()

    print "Load mesh"
    m = taglut.Mesh(inputFile)
    print "Load scalar function"
    mmap = taglut.MeshMap(functionFile)
    mmap.setMesh(m)
    print "Correct scalar function (removing flat edges)"
    mmap.removeFlatEdges()
    print "Compute the corresponnding vector field"
    vfield = taglut.VectorField(mmap, m)

    print "Compute the n-loops"
    nl = taglut.NLoopFactory.getNLoopsAsPLPathFromSaddlePoints(m, mmap, vfield, angles)

    if (infos):
        for p in nl:
            print "Found an n-loop. Arity:", p.getNbPaths()
            for pp in p.getPaths():
                print " found a path with", len(pp), "points"

    print "Cut mesh using n-loops..."
    plCut = taglut.MeshPLCut()
    m2 = plCut.cutMesh(m, nl)

    if quads:
        print "Split extrema discs..."
        m3 = plCut.cutExtremaDiscs(m2, mmap.getLocalExtrema())
    else:
        m3 = m2

    if not cylinders:
        print "Split cylinders..."
        m4 = plCut.cutMeshInQuadranglesFromCylindricalTiling(m3)
    else:
        m4 = m3

    if (infos):
        print m4.getInfos()
    if (display):
        print "Display..."
        taglut.Display3D.displayMesh(sizex, sizey, "Display", m4, False, True)
    if (outputFile != ""):
        print "Save mesh (" + outputFile + ")..."
        m3.save(outputFile)
    if (outputFileParam != ""):
        print "Parameterization"
        mtiling = taglut.MTiling(m4)
        if quads:
            mtiling.adjustQuads()
        mapping = mtiling.parameterize()
        print "Save mapping (" + outputFileParam + ")..."
        mapping.save(outputFileParam)

if __name__ == "__main__":
    main()
