/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2010 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal / IMATI-CNR
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "NLoopFactory.h"
#include "VectorField.h"
#include "PLPath.h"
#include "NLoop.h"
#include "PreNLoop.h"

using namespace Taglut;



Pre3Loop NLoopFactory::getPre3LoopFromSaddlePoint(const Point3D & point,
						  const Mesh & mesh,
						  const MeshMap & mMap, const VectorField & vf) {
  return Pre3Loop(point, mesh, mMap, vf);
}




std::vector<PreNLoop> NLoopFactory::getPreNLoopsFromSaddlePoints(const Mesh & mesh,
								 const MeshMap & mMap, const VectorField & vf,
								 bool splitForNLoop,
								 bool merge) {
  std::vector<PreNLoop> result;

  // first build the pre-n-loops
  std::vector<std::pair<VertexID, unsigned int> > sps = mMap.getSaddlePoints();
  for(std::vector<std::pair<VertexID, unsigned int> >::const_iterator sp = sps.begin(); sp != sps.end(); ++sp) {
    if ((*sp).second == 2) {
      try {
	Pre3Loop pre3loop = getPre3LoopFromSaddlePoint(mesh.point((*sp).first), mesh, mMap, vf);
	if ((merge || splitForNLoop) && (result.size() != 0)) {
	  bool inserted = false;
	  std::vector<PreNLoop>::iterator pnlInserted;
	  for(std::vector<PreNLoop>::iterator pnl = result.begin(); pnl != result.end(); ++pnl)
	    if ((*pnl).intersects(pre3loop, mMap)) {
	      if (!inserted) {
		(*pnl).addPre3Loop(pre3loop);
		inserted = true;
		pnlInserted = pnl;
	      }
	      else {
		(*pnlInserted).addPre3Loops((*pnl).getPre3Loops());
		(*pnl).clear();
	      }
	    }
	  if (!inserted) {
	    result.push_back(PreNLoop(pre3loop));
	  }
	}
	else {
	  result.push_back(PreNLoop(pre3loop));
	}
      }
      catch (Exception e) {
	std::cout << "Warning: error during pre-n-loop reconstruction..." << std::endl;
	std::cout << " Exception: " << e << std::endl;
      }
    }
    else {
      std::cout << "Warning: a non regular saddle point has been detected. Cannot build an n-loop on it." << std::endl;
    }
  }
  if (splitForNLoop) {
    std::vector<PreNLoop> result2;
    for(std::vector<PreNLoop>::iterator pnl = result.begin(); pnl != result.end(); ++pnl)
      if (!(*pnl).empty()) {
	if ((*pnl).isWellOrdered(mMap)) {
	  result2.push_back(*pnl);
	}
	else {
	  std::vector<PreNLoop> sp = (*pnl).split(mMap);
	  result2.insert(result2.end(), sp.begin(), sp.end());
	}
      }
    result = result2;
  }
  else {
    std::vector<PreNLoop> result2;
    for(std::vector<PreNLoop>::iterator pnl = result.begin(); pnl != result.end(); ++pnl)
      if (!(*pnl).empty())
	result2.push_back(*pnl);
    result = result2;
  }

  return result;
}


std::vector<NLoop> NLoopFactory::getNLoopsFromSaddlePoints(Mesh & mesh,
							   const MeshMap & mMap, const VectorField & vf, bool adjustByAngles) {
  // first compute the pre-n-loops using the scalar function
  std::vector<PreNLoop> preNLoops = getPreNLoopsFromSaddlePoints(mesh, mMap, vf);
  std::vector<NLoop> result;

  for(std::vector<PreNLoop>::const_iterator preNLoop = preNLoops.begin(); preNLoop != preNLoops.end(); ++preNLoop) {
    result.push_back((*preNLoop).getNLoop(mesh, mMap, vf, adjustByAngles));
  }
  return result;
}

std::vector<PLNLoop> NLoopFactory::getPLNLoopsFromSaddlePoints(Mesh & mesh,
							   const MeshMap & mMap, const VectorField & vf, bool adjustByAngles) {
  // first compute the pre-n-loops using the scalar function
  std::vector<PreNLoop> preNLoops = getPreNLoopsFromSaddlePoints(mesh, mMap, vf);
  std::vector<PLNLoop> result;

  for(std::vector<PreNLoop>::const_iterator preNLoop = preNLoops.begin(); preNLoop != preNLoops.end(); ++preNLoop) {
    result.push_back((*preNLoop).getPLNLoop(mesh, mMap, vf, adjustByAngles));
  }
  return result;
}

std::vector<PLPath> NLoopFactory::getNLoopsAsPLPathFromSaddlePoints(Mesh & mesh, const MeshMap & mMap,
								    const VectorField & vf, bool adjustByAngles) {
  std::vector<PLNLoop> nloops = getPLNLoopsFromSaddlePoints(mesh, mMap, vf, adjustByAngles);
  std::vector<PLPath> result;

  for(std::vector<PLNLoop>::const_iterator nloop = nloops.begin(); nloop != nloops.end(); ++nloop) {
    const std::vector<PLPath> & paths = (*nloop).getPaths();
    result.insert(result.end(), paths.begin(), paths.end());
  }

  return result;
}
