/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2010 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal / IMATI-GE CNR
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include "CylinderParameterization.h"
#include "Mesh.h"
#include "MeshManipulator.h"

using namespace Taglut;


void TwoCirclesBorderParameterization::parameterizeBorder3D(Mesh & mesh, Mesh & resultMesh) {
  if ((bdPoint[0][0] == bdPoint[0][1]) ||
      (bdPoint[1][0] == bdPoint[1][1]))
    throw Exception("parametrizeBorder3D(2): no border points selected");

  MeshManipulator mManip(mesh);
  std::vector<VertexID> bd[2] = { mManip.getBoundaryFromBPoint(bdPoint[0][0]), mManip.getBoundaryFromBPoint(bdPoint[1][0])};

  if (adjust)
    adjustParameters(mesh, bd);

  for(unsigned char b = 0; b != 2; ++b) {
    const unsigned char b2 = b == 0 ? 1 : 0;
    // first compute the number of points on each side
    unsigned int nbFirstBow = 0;
    if (bd[b].front() != bdPoint[b][0])
      throw Exception("parametrizeBorder3D(2): incoherent boundary");
    for(std::vector<VertexID>::const_iterator bdb = bd[b].begin(); bdb != bd[b].end(); ++bdb, ++nbFirstBow) {
      if (((*bdb) == bdPoint[b2][0]) ||
	  ((*bdb) == bdPoint[b2][1]))
	throw Exception("parametrizeBorder3D(2): incoherent boundaries");
      if ((*bdb) == bdPoint[b][1])
	break;
    }
    if ((nbFirstBow == bd[b].size() - 1) && (bd[b].back() != bdPoint[b][1]))
      throw Exception("parametrizeBorder3D(2): incoherent boundary (missing point)");

    // then adjust the location of the boundary points in the result mesh
    unsigned int i = 0;
    double angle = 0.;
    double step = M_PI / nbFirstBow;
    bool inverse = false;
    // adjust the order
    {
      assert(bd[b].size() > 1);
      const Triangle & t = mesh.findTriangle(bd[b].front(), bd[b][1]);
      const bool o = t.sameOrder(bd[b].front(), bd[b][1]);
      if (((b == 0) && o) ||
	  ((b == 1) && !o))
	inverse = true;
    }
    if (inverse) {
      angle = 2 * M_PI;
      step = - step;
    }
    for(std::vector<VertexID>::const_iterator bdb = bd[b].begin(); bdb != bd[b].end(); ++bdb, ++i, angle += step) {
      resultMesh.point(*bdb).setX(radius * cos(angle));
      resultMesh.point(*bdb).setY(radius * sin(angle));
      resultMesh.point(*bdb).setZ(b * length);
      if (*bdb == bdPoint[b][1])
	step = (inverse ? -1 : 1 ) * M_PI / (bd[b].size() - nbFirstBow);
    }

  }
  // reset the boundary points
  bdPoint[0][0] = bdPoint[0][1] = bdPoint[1][0] = bdPoint[1][1] = 0;
}

void CylinderParameterization::initSolutionVector(OpenNLSolver & solver, unsigned char i) {
  float value = 0.;
  if (i == 2)
    value = length * .5;

  for(VertexID v = 0; v != solver.getSize(); ++v)
    solver.setVariable(v, i);

}

void CylinderParameterization::projectMeshOnCylinder(Mesh & mesh) {
  for(Mesh::point_iterator p = mesh.point_begin(); p != mesh.point_end(); ++p)
    if (!((*p).getIsBoundary())) {
      const Coord3D c(0., 0., (*p).getZ());
      (*p).setCoords(c + ((*p) - c) / (*p).distance(c) * radius);
    }
}

void TwoCirclesBorderParameterization::adjustParameters(Mesh & mesh, const std::vector<VertexID> * bd) {
  MeshManipulator mManip(mesh);
  radius = (mesh.getLength(bd[0]) + mesh.getLength(bd[0])) / (4 * M_PI);
  length = mManip.getGeodesicDistance(bd[0], bd[1]);
}
