/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2010 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal / IMATI-GE CNR
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#ifndef CYLINDER_PARAMETERIZATION
#define CYLINDER_PARAMETERIZATION

#include "BarycentricMapping.h"
#include "OpenNLSolver.h"
#include "IDTypes.h"
#include "Mesh.h"

namespace Taglut {

  class TwoCirclesBorderParameterization {
  protected:
    /** boundary points */
    VertexID bdPoint[2][2];

    /** radius of the boundary circles */
    double radius;
    /** length of the cylinder */
    double length;
    /** true if length and radius has to be adjusted according to the input mesh */
    bool adjust;

    void parameterizeBorder3D(Mesh & mesh, Mesh & resultMesh);

    void parameterizeBorder(Mapping2D3D & ) {
      throw Exception("TwoCirclesBorderParameterization: only 3D description");
    }

    void adjustParameters(Mesh & mesh, const std::vector<VertexID> * bd);

    inline void initValues(double radius_t = 1., double length_t = 2.) {
      bdPoint[0][0] = bdPoint[0][1] = bdPoint[1][0] = bdPoint[1][1] = 0;
      radius = radius_t;
      length = length_t;
      adjust = false;
    }
    inline void setAdjust(bool a) {
      adjust = a;
    }

    inline void setBPoints(VertexID bdPoints1[2], VertexID bdPoints2[2]) {
      bdPoint[0][0] = bdPoints1[0];
      bdPoint[0][1] = bdPoints1[1];
      bdPoint[1][0] = bdPoints2[0];
      bdPoint[1][1] = bdPoints2[1];
    }
  public:
    TwoCirclesBorderParameterization() {
      initValues();
    }

  };

  /**
     @class CylinderParameterization
     @author Jean-Marie Favreau (LIMOS / Blaise Pascal Univ. / IMATI-GE)
     @brief Implementation of a barycentric mapping of a tubular surface by cylinder
  */
  class CylinderParameterization : public MeanValuesParameterization<TwoCirclesBorderParameterization> {
  protected:


    virtual void initSolutionVector(OpenNLSolver & solver, unsigned char i);

    void projectMeshOnCylinder(Mesh & mesh);

  public:
    // default constructor
    CylinderParameterization(double radius_t = 1., double length_t = 2.) {
      initValues(radius_t, length_t);
    }

    /** compute the parameterization given two points on each boundary
	to define the border parameterization
	@param mesh The mesh to be parameterized (required: it have to be a cylinder)
	@param bdPoints1 Two points in the first boundary
	@param bdPoints2 Two points in the second boundary
	@param project Project the resulting mesh on the cylinder
	@param adjust Adjust the radius and length of the resulting structure according to the average of the length of the boundaries, and the length of the shortest path between the two boundaries

    */
    inline Mesh parameterize3DFixedPoints(Mesh & mesh, VertexID bdPoints1[2], VertexID bdPoints2[2], bool project = false, bool adjust_t = true) {
      setAdjust(adjust_t);
      // select points on boundary
      setBPoints(bdPoints1, bdPoints2);

      Mesh r = parameterize3D(mesh);
      if (project)
	projectMeshOnCylinder(r);
      // compute the parameterization
      (*this).adjust = false;
      return r;
    }


  };

}

#endif
