/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2008 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef MESH_MAP
#define MESH_MAP

#include <vector>
#include "IDTypes.h"
#include "Coord3D.h"
#include "Point3D.h"
#include "FileManipulator.h"



namespace Taglut {
  class Mesh;
  class MeshManipulator;
  class MeshPart;
  class PLPath;
  class PointOnEdge;


  /**
     @class MeshMapT

     @author Jean-Marie Favreau (LIMOS/Blaise Pascal Univ.)
     @date 2008-02-24
     @brief Mesh map: a double value is associated for each point of the mesh
  */
  template <typename T>
  class MeshMapT : public FileManipulator {
  protected:
    /**
       Mesh that is the support of the current map
    */
    Mesh * mesh;

    /**
       Array of values associated for each point of the mesh
    */
    T * values;

    /**
       Number of values (should be equal to the number of points on the mesh)
    */
    unsigned int nbValues;

    /**
       Adding the supported saving and loading formats
    */
    inline void addFormats() {
      addFormat("txt");
      addFormat("usr");
      addFormat("hdr");
      addFormat("dcm");
      addSaveFormat("vtk");
    }

    /**
       For the given point, compute in its neighbourhood the number of values' changes (up or down)
       according to the value of p (that may be given in the second parameter
    */
    unsigned int getNeighbourArity(const Point3D & p, T * curValue = NULL) const;

  public:

    /**
       Empty constructor
    */
    MeshMapT();

    /**
       Default constructor
    */
    MeshMapT(Mesh & mesh, T value = 1);

    /**
       Create a meshmap on the mesh from a part of an original mesh, and the corresponding meshmap
       @param mesh The submesh
       @param meshPart The description of the inclusion (with the mapping of vertices)
       @param refMeshMap The reference MeshMapT where the values will be taken
    */
    MeshMapT(Mesh & mesh, const MeshPart & meshPart, const MeshMapT<T> & refMeshMap);

    /**
       build a function defined by the plane (one side with negative values, one side with
       positive values, and value defined by distance to the plane)
    */
    MeshMapT(Mesh & mesh, const Plane3D & plane);


    /**
       Copy constructor
    */
    MeshMapT(const MeshMapT<T> & mMap);

    /**
       Load values from a file, without connecting it to a mesh
    */
    MeshMapT(const std::string & filename);

    /**
       Load values from a file
    */
    MeshMapT(Mesh & mesh, const std::string & filename);

#ifndef SWIG
    /**
       Load values from a file
    */
    MeshMapT(Mesh & mesh, const char * filename);
#endif

    /** "copy" constructor using a new mesh */
    MeshMapT(const MeshMapT<T> & mMap, Mesh & mesh);

    /**
       Destructor
    */
    ~MeshMapT() {
      if (values != NULL)
	delete [] values;
    }

    /**
       Load Mesh from file
    */
    virtual void loadAltFormat(const std::string & fileName, const std::string & altFormat = "", const std::string & object = "");

    /**
       Save Mesh to file
    */
    virtual void saveAltFormat(const std::string & fileName, const std::string & format = "", const std::string & objectName  = "") const;

    /**
       Save values to file (ascii format)
    */
    void saveTXT(const std::string & fileName) const;

    /**
       Save values (and mesh) to file (vtk ascii format)
    */
    void saveVTK(const std::string & fileName) const;

    /**
       Load values from a file (ascii format)
    */
    void loadTXT(const std::string & fileName);

    /**
       Load MeshMap from file (volumic image, like analyze or dicom). Assuming that a mesh has been associated to
       the current mesh map, the values of the meshmap are computed using the coordinates of the vertices on the image
    */
    void loadFromVolumicImage(const std::string & fileName, double ratio = 1.);


    /**
       Set the mesh that will be manipulated
    */
    inline const Mesh & getMesh() const { return *mesh; }


    /**
       Set the mesh that will be manipulated
    */
    void setMesh(Mesh & mesh);

    /**
       Set the mesh that will be manipulated, with an initial value
    */
    void setMesh(Mesh & mesh, T value);

    /**
       Update the mesh by adding the possible missing points
       \see addMissingValuesUsingMean
    */
    void updateMesh(Mesh & mesh);

    /**
       return the number of values in the current mesh map
    */
    inline VertexID getNbValues() const {
      return nbValues;
    }

    /**
       Accessor for the map value associated to a given point defined by \p id
    */
    inline const T & getValue(VertexID id) const {
      if (values == NULL)
	throw Exception("getValue(1): values not defined.");
      if (id >= nbValues)
	throw Exception("getValue(1): id out of bounds.");
      return values[id];
    }

    /**
       set the values according to the flags
    */
    void setValuesFromFlags();

    /**
       set the values according to lengths
    */
    void setValuesFromLengths(const MeshManipulator & mManip);

#ifndef SWIG
    /**
       Accessor for the map value associated to a given point defined by \p id
    */
    inline const T & operator[] (VertexID id) const {
      if (values == NULL)
	throw Exception("operator[](1): values not defined.");
      if (id >= nbValues)
	throw Exception("operator[](1): id out of bounds.");
      return values[id];
    }

    /**
       Accessor for the map value associated to a given point defined by \p id.
    */
    inline T & operator[] (VertexID id) {
      if (values == NULL)
	throw Exception("operator[](1): values not defined.");
      if (id >= nbValues)
	throw Exception("operator[](1): id out of bounds.");
      return values[id];
    }
#endif

    /**
       Modifier for the map value associated to a given point defined by \p id
    */
    void setValue(VertexID id, T value);

    /**
       init all the values using the given one
    */
    void initValues(T value);

    /**
       Given a center defined by its point's id \p id, an initial value (\p epsilon),
       and a scale value (\p alpha), define value for each point of the mesh
    */
    void setValuesFromCenter(VertexID id, double alpha = 1.0, double epsilon = 1e-1);

    /**
       Given a center defined by a direction \p direction, an initial value (\p epsilon),
       and a scale value (\p alpha), define value for each point of the mesh
    */
    void setValuesFromDirection(const Direction & direction, double alpha = 1.0, double epsilon = 1e-1);

    /**
       return the maximum value of the mapping
    */
    T getMaxValue() const;

    /**
       return the minimum value of the mapping
    */
    T getMinValue() const;

    /**
       Using a Gaussian filter to smooth the values
    */
    void smoothValues();

    /**
       Remove flat edges by introducing small values
    */
    void removeFlatEdges();

    /**
       Apply a log function filter to modify the values
    */
    void applyLog();

    /**
       If points as been added to the mesh, and assuming that the insertion has been done at the end
       of the list, update the values of this MeshMapT by adding the missing values and estimate it using
       the mean of its neighbours in the original mesh.
    */
    void addMissingValuesUsingMean();

    /**
       return the value using a linear approximation along the edge
    */
    T getValue(const PointOnEdge & point) const;

    /**
       Given a point on an edge, compute the associated connected component of the level set on the function
       defined by the current mesh map.
    */
    PLPath computeLevelSetCC(const PointOnEdge & point) const;

    /**
       Given a value, compute the corresponding level sets.
    */
    std::vector<PLPath> computeLevelSet(double value) const;

    /**
       Given an edge (i.e. 2 points) and a scalar value, return the corresponding point on edge, or
       throws an exception if the value is not between the values of p1 and p2
     */
    PointOnEdge getPointOnEdgeFromValueAndEdge(VertexID p1, VertexID p2, T value) const;

    /**
       given a path (that is probably an integral line) and a value, compute a point close to the path with
       the given value. This point is on the path if possible, or in an edge that is connected to the
       path with one triangle.
     */
    PointOnEdge getPointOnEdgeFromValueAndPath(const PLPath & path, double value) const;

    /**
       given a path (that is probably an integral line) and a value, compute the connected
       component of level set associated to the given value that cross the given path
     */
    PLPath computeLevelSetFromValueAndPath(const PLPath & path, double value) const;

    /**
       Given a point on an edge, an adjacent triangle and the level set value associated to the point,
       compute the next edge point on the triangle that corresponds to the level set.
    */
    PointOnEdge getNextPointOnLevelSet(const PointOnEdge & point, TriangleID t, double lsvalue) const;

    /**
       Given a point (that is a vertex) and the level set value associated to the point,
       compute a possible next edge point that corresponds to the level set.
    */
    PointOnEdge getPossibleNextPointOnLevelSet(const PointOnEdge & point, double lsvalue) const;

    /**
       Given a point (that is a vertex), the level set value associated to the point, and a previously found point,
       compute the next edge point that corresponds to the level set.
    */
    PointOnEdge getOtherPossibleNextPointOnLevelSet(const PointOnEdge & point, double lsvalue, const PointOnEdge & pointfirst) const;

    /**
       return true if the given points has the same value
    */
    inline bool isFlatEdge(VertexID p1, VertexID p2) const {
      return values[p1] == values[p2];
    }

    /**
       Return the saddle points (and the corresponding complexity of the saddle) of the scalar function described by the mesh map.
    */
    std::vector<std::pair<VertexID, unsigned int> > getSaddlePoints() const;

    /**
       Return local extrema
    */
    std::vector<VertexID> getLocalExtrema() const;

    /**
       Return the number of flat edges
    */
    VertexID getNbFlatEdges() const;

    /**
       return true if the given point is a saddle point
    */
    bool isSaddlePoint(VertexID p) const;


    /**
       return a map that contains for each point the absolute difference between the value
       of the current mesh map and the given value
    */
    MeshMapT<T> distanceMap(T value) const;

    /**
       return the minimum (non zero) difference on edge
     */
    T getMinimumDifferenceOnEdge() const;

    /**
       return the area of the surface contained between the two values
     */
    double getArea(T minValue, T maxValue) const;

    /**
       return the area of the surface in one side of the value. If \p sup = true, the superior side is choosen.
     */
    double getArea(T value, bool sup) const;
  };

  /** a MeshMap with doubles */
  typedef MeshMapT<double> MeshMap;
}
#define MESH_MAP_DEFINED

#include "Mesh.h"
#include "MeshManipulator.h"
#include "MeshPart.h"
#include "FileExceptions.h"
#include "PointOnEdge.h"
#include "PLPath.h"

namespace Taglut {

  template <typename T>
  MeshMapT<T>::MeshMapT() {
    addFormats();
    mesh = NULL;
    values = NULL;
    nbValues = 0;
  }

  template <typename T>
  MeshMapT<T>::MeshMapT(Mesh & mesh_l, T value) {
    addFormats();
    (*this).mesh = &mesh_l;
    nbValues = mesh_l.getNbPoints();
    values = new T[nbValues];
    T * v = values;
    for(unsigned int i = 0; i < nbValues; ++i, ++v)
      *v = value;
  }

  template <typename T>
  MeshMapT<T>::MeshMapT(Mesh & mesh_l, const Plane3D & plane) {
    addFormats();
    (*this).mesh = &mesh_l;
    nbValues = mesh_l.getNbPoints();
    values = new T[nbValues];
    T * v = values;

    for(Mesh::const_point_iterator p = mesh_l.point_begin(); p != mesh_l.point_end(); ++p, ++v)
      *v = plane.signedDistance(*p);
  }


  template <typename T>
  MeshMapT<T>::MeshMapT(const MeshMapT<T> & mMap, Mesh & mesh_l) : FileManipulator(mMap) {
    addFormats();
    (*this).mesh = &mesh_l;
    if ((mMap.values != NULL) && (nbValues != 0)) {
      nbValues = mMap.nbValues;
      values = new T[nbValues];
      T * v = values;
      T * oV = mMap.values;
      for(unsigned int i = 0; i < nbValues; ++i, ++v, ++oV)
	*v = *oV;
    }
    else {
      nbValues = 0;
      values = NULL;
    }
    addMissingValuesUsingMean();
  }

  template <typename T>
  MeshMapT<T>::MeshMapT(const MeshMapT<T> & mMap) : FileManipulator(mMap) {
    addFormats();
    if (mMap.mesh != NULL)
      mesh = mMap.mesh;
    nbValues = mMap.nbValues;
    if ((mMap.values != NULL) && (nbValues != 0)) {
      values = new T[nbValues];
      T * v = values;
      T * oV = mMap.values;
      for(unsigned int i = 0; i < nbValues; ++i, ++v, ++oV)
	*v = *oV;
    }
    else {
      nbValues = 0;
      values = NULL;
    }
  }


  template <typename T>
  MeshMapT<T>::MeshMapT(const std::string & filename) {
    addFormats();
    (*this).mesh = NULL;
    nbValues = 0;
    values = NULL;
    load(filename);
  }

  template <typename T>
  MeshMapT<T>::MeshMapT(Mesh & mesh_l, const std::string & filename) {
    addFormats();
    (*this).mesh = &mesh_l;
    nbValues = mesh_l.getNbPoints();
    values = new T[nbValues];
    load(filename);
  }

  template <typename T>
  MeshMapT<T>::MeshMapT(Mesh & mesh_l, const char * filename) {
    addFormats();
    (*this).mesh = &mesh_l;
    nbValues = mesh_l.getNbPoints();
    values = new T[nbValues];
    load(filename);
  }

  template <typename T>
  MeshMapT<T>::MeshMapT(Mesh & mesh_l, const MeshPart & meshPart, const MeshMapT<T> & refMeshMapT) {
    addFormats();
    (*this).mesh = &mesh_l;
    nbValues = mesh_l.getNbPoints();
    values = new T[nbValues];
    if (mesh_l.getNbPoints() != meshPart.getNbPoints())
      throw Exception("MeshMapT(): wrong number of points in the meshpart.");

    const VertexID maxID = meshPart.getMesh().getNbPoints();
    if (maxID != refMeshMapT.getNbValues())
      throw Exception("MeshMapT(): wrong number of points in the original mesh.");

    T * v = values;
    VertexID mapID = 0;

    for(VertexID i = 0; i != maxID; ++i)
      if (meshPart.hasPoint(i)) {
	*v = refMeshMapT[i];
	++v;
	++mapID;
	if (mapID == nbValues)
	  break;
      }
  }


  template <typename T>
  void MeshMapT<T>::loadAltFormat(const std::string & fileName, const std::string & altFormat, const std::string &) {
    if ((altFormat == "txt") || (altFormat == "usr"))
      loadTXT(fileName);
    else if ((altFormat == "dcm") || (altFormat == "hdr"))
      loadFromVolumicImage(fileName);
    else
      throw Exception("Unknown format");
  }

  template <typename T>
  void MeshMapT<T>::saveAltFormat(const std::string & fileName, const std::string & format, const std::string &) const {
    if ((format == "txt") || (format == "usr"))
      saveTXT(fileName);
    else if (format == "vtk")
      saveVTK(fileName);
    else
      throw Exception("Unknown format");
  }

  template <typename T>
  void MeshMapT<T>::saveTXT(const std::string & fileName) const {
    std::ofstream outfile(fileName.c_str(), std::ios::out);

    if (!outfile.is_open())
      throw ExceptionCannotOpenOutputFile();

    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      outfile << *v << std::endl;

    outfile.close();
  }

  template <typename T>
  void MeshMapT<T>::saveVTK(const std::string & fileName) const {
    std::ofstream outfile(fileName.c_str(), std::ios::out);

    if (!outfile.is_open())
      throw ExceptionCannotOpenOutputFile();

    // save mesh
    outfile << "# vtk DataFile Version 3.0" << std::endl;

    outfile << "vtk output" << std::endl << "ASCII" << std::endl << "DATASET POLYDATA" << std::endl;
    outfile << "POINTS " << (*mesh).getNbPoints() << " float" << std::endl;

    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p)
      outfile << (*p).getX() << " " << (*p).getY() << " " << (*p).getZ() << std::endl;

    outfile << "POLYGONS " << (*mesh).getNbTriangles() << " " << ((*mesh).getNbTriangles() * 4) << std::endl;

    for(Mesh::const_triangle_iterator t = (*mesh).triangle_begin(); t != (*mesh).triangle_end(); ++t)
      outfile << "3 " << (*t).getP1() << " " << (*t).getP2() << " " << (*t).getP3() << std::endl;

    // save values
    outfile << "POINT_DATA " << nbValues << std::endl;
    outfile << "SCALARS values float" << std::endl;
    outfile << "LOOKUP_TABLE default" << std::endl;
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      outfile << *v << std::endl;

    outfile.close();
  }

  template <typename T>
  void MeshMapT<T>::loadTXT(const std::string & fileName) {
    std::ifstream infile(fileName.c_str());

    if (!infile.is_open())
      throw ExceptionFileNotFound();

    if (values != NULL)
      delete [] values;

    if (mesh != NULL) {
      nbValues = (*mesh).getNbPoints();
      values = new T[nbValues];

      VertexID cpt = 0;
      T * v = values;
      while(!infile.eof()) {
	infile >> *v;
	++v;
	++cpt;
	if (cpt == nbValues)
	  break;
      }
      if (cpt != nbValues)
	throw ExceptionDataNotFoundInFile();
    }
    else {
      // load file
      std::vector<T> lv;
      T v;
	infile >> v;
      while(!infile.eof()) {
	lv.push_back(v);
	infile >> v;
      }

      // copy it on the structure
      nbValues = lv.size();
      values = new T[nbValues];
      T * vv = values;
      for(typename std::vector<T>::const_iterator lvv = lv.begin(); lvv != lv.end(); ++lvv, ++vv)
	*vv = *lvv;

    }

    infile.close();
  }


  template <typename T>
  void MeshMapT<T>::loadFromVolumicImage(const std::string & fileName, double ratio) {
    if (mesh == NULL) {
      throw Exception("loadFromVolumicImage(2): the mesh has to be defined to load data from a volumic image.");
    }
    if (values != NULL)
      delete [] values;
    nbValues = (*mesh).getNbPoints();
    values = new T[nbValues];

    const CImg<double> img(fileName.c_str());

    T *v = values;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++v, ++p) {
      unsigned int x = floor((*p).getX());
      unsigned int y = floor((*p).getY());
      unsigned int z = floor((*p).getZ() / ratio);
      if (x != (*p).getX()) {
	*v = (img(x, y, z) + img(x + 1, y, z)) / 2;
      }
      else if (y != (*p).getY()) {
	*v = (img(x, y, z) + img(x, y + 1, z)) / 2;
      }
      else {
	assert(z != (*p).getZ());
	*v = (img(x, y, z) + img(x, y, z + 1)) / 2;
      }
    }

  }

  template <typename T>
  void MeshMapT<T>::setMesh(Mesh & mesh_l, T value) {
    if (values != NULL)
      delete [] values;
    (*this).mesh = &mesh_l;
    nbValues = mesh_l.getNbPoints();
    values = new T[nbValues];
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      *v = value;
  }

  template <typename T>
  void MeshMapT<T>::setMesh(Mesh & mesh_l) {
    (*this).mesh = &mesh_l;
  }


  template <typename T>
  void MeshMapT<T>::setValue(VertexID id, T value) {
    if (values == NULL)
      throw Exception("setValue(2): values not defined.");
    if (id >= nbValues)
      throw Exception("setValue(2): id out of bounds.");
    values[id] = value;
  }

  template <typename T>
  void MeshMapT<T>::initValues(T value) {
    if (values == NULL)
      throw Exception("initValues(): values not defined.");
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      *v = value;
  }

  template <typename T>
  void MeshMapT<T>::setValuesFromFlags() {
    if (mesh == NULL)
      throw Exception("setValuesFromFlags(): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("setValuesFromFlags(): values not defined.");
    VertexID i = 0;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p, ++i)
      values[i] = (T)((*p).getFlag());
  }

  template <typename T>
  void MeshMapT<T>::smoothValues() {
    if (mesh == NULL)
      throw Exception("smoothValues(): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("smoothValues(): values not defined.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("smoothValues(): mesh has been modified (bad number of values).");
    VertexID i = 0;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p, ++i) {
      unsigned int nbNb = 2;
      values[i] += values[i];
      for (std::deque<VertexID>::const_iterator nb = (*p).getNeighbours().begin(); nb != (*p).getNeighbours().end(); ++nb, ++nbNb)
	values[i] += values[*nb];
      values[i] /= nbNb;
    }
  }

  template <typename T>
  void MeshMapT<T>::removeFlatEdges() {
    if (mesh == NULL)
      throw Exception("removeFlatEdges(): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("removeFlatEdges(): values not defined.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("removeFlatEdges(): mesh has been modified (bad number of values).");

    // first compute the smallest difference between points
    double value = std::numeric_limits<double>::max();
    VertexID i = 0;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p, ++i) {
      const double v1 = values[i];
      for (std::deque<VertexID>::const_iterator nb = (*p).getNeighbours().begin(); nb != (*p).getNeighbours().end(); ++nb) {
	const double diff = fabs(values[*nb] - v1);
	if ((diff != 0.) && (diff < value))
	  value = diff;
      }
    }
    // then modify the values using this small value
    value /= nbValues + 1;
    assert(value != 0.);
    double * cValue = values;
    for(VertexID ii = 0; ii != nbValues; ++ii, ++cValue) {
#ifndef NDEBUG
      const double previous = *cValue;
#endif
      (*cValue) += ii * value;
#ifndef NDEBUG
      assert((ii == 0) || (*cValue != previous));
#endif
    }

  }


  template <typename T>
  void MeshMapT<T>::setValuesFromLengths(const MeshManipulator & mManip) {
    if (nbValues != mManip.getMesh().getNbPoints())
      throw Exception("setValuesFromLengths(1): error (bad number of values in MeshManipulator ).");
    for(VertexID i = 0; i < nbValues; ++i)
      values[i] = (T)mManip.getLength(i);

  }


  template <typename T>
  void MeshMapT<T>::setValuesFromCenter(VertexID id, double alpha, double epsilon) {
    if (mesh == NULL)
      throw Exception("setValuesFromCenter(3): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("setValuesFromCenter(3): values not defined.");
    if (id >= nbValues)
      throw Exception("setValuesFromCenter(3): id out of bounds.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("setValuesFromCenter(3): mesh has been modified (bad number of values).");
    MeshManipulator mManip(*mesh);

    mManip.computeDijkstra(id);
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      *v = alpha * mManip.getLength(i) + epsilon;
  }

  template <typename T>
  void MeshMapT<T>::setValuesFromDirection(const Direction & direction, double alpha, double epsilon) {
    if (mesh == NULL)
      throw Exception("setValuesFromDirection(3): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("setValuesFromDirection(3): values not defined.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("setValuesFromDirection(3): mesh has been modified (bad number of values).");

    if (direction == DNone)
      return;

    MeshManipulator mManip(*mesh);


    Line3D line((*mesh).getIsobarycenter(), direction);
    unsigned int id = (*mesh).getLastIntersection(line);

    if (id == (*mesh).getNbPoints())
      std::cout << "Warning: cannot find points on the axis." << std::endl;
    else
      setValuesFromCenter(id, alpha, epsilon);
  }

  template <typename T>
  T MeshMapT<T>::getMaxValue() const {
    T result = std::numeric_limits<T>::min();
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      if (*v > result)
	result = *v;
    return result;
  }

  template <typename T>
  T MeshMapT<T>::getMinValue() const {
    T result = std::numeric_limits<double>::max();
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      if (*v < result)
	result = *v;
    return result;
  }

  template <typename T>
  void MeshMapT<T>::applyLog() {
    T * v = values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      *v = log(*v);
  }

  template <typename T>
  void MeshMapT<T>::addMissingValuesUsingMean() {
    if (mesh == NULL)
      throw Exception("addMissingValuesUsingMean(): mesh not defined in MeshMap");
    if ((*mesh).getNbPoints() == nbValues)
      return;

    VertexID nbValuesOld = nbValues;
    T * valuesOld = values;
    nbValues = (*mesh).getNbPoints();
    values = new T[nbValues];
    T * v = values;
    T * vold = valuesOld;
    // first copy the values
    for(VertexID i = 0; i < nbValuesOld; ++i, ++v, ++vold)
      *v = *vold;

    // then update the new one using the mean of its neighbours
    for(VertexID i = nbValuesOld; i < nbValues; ++i, ++v) {
      T r = 0.0;
      unsigned int nbNbOld = 0;
      for(std::deque<VertexID>::const_iterator nb = (*mesh).point(i).getNeighbours().begin(); nb != (*mesh).point(i).getNeighbours().end(); ++nb)
	if (*nb < nbValuesOld) {
	  ++nbNbOld;
	  r += valuesOld[*nb];
	}
      if (nbNbOld != 0)
	r /= nbNbOld;
      *v = r;
    }
    delete [] valuesOld;
  }

  template <typename T>
  void MeshMapT<T>::updateMesh(Mesh & mesh_l) {
    (*this).mesh = &mesh_l;
    addMissingValuesUsingMean();
  }

  template <typename T>
  PLPath MeshMapT<T>::computeLevelSetCC(const PointOnEdge & point) const {
    if (mesh == NULL)
      throw Exception("computeLevelSetCC(1): mesh not defined in MeshMap");
    PLPath path(*mesh);
    path.addPointAtEnd(point);
    T lsvalue = point.getValueFromPLFunction(values[point.getFirstVertex()], values[point.getSecondVertex()]);
    bool not_closed = true;
    bool firstSide = true;

    while(not_closed) {
      assert(fabs(getValue(path.back()) - getValue(point)) < 1e-4);
      bool reverse = false;
      if (path.back().isVertex()) {
	if (path.size() == 1) {
	  PointOnEdge nextp = getPossibleNextPointOnLevelSet(path.back(), lsvalue);
	  path.addPointAtEnd(nextp);
	}
	else {
	  PointOnEdge nextp = getPossibleNextPointOnLevelSet(path.back(), lsvalue);
	  const PointOnEdge predp = path[path.size() - 2];
	  if (nextp == predp) {
	    try {
	      nextp = getOtherPossibleNextPointOnLevelSet(path.back(), lsvalue, predp);
	    }
	    catch (Exception e) {
	      if (firstSide) {
		firstSide = false;
		path.reversePath();
		reverse = true;
	      }
	      else
		not_closed = false;
	    }
	  }
	  if (!reverse) {
	    path.addPointAtEnd(nextp);
	    if (path.front().inSameEdge(path.back()) && (path.size() > 2)) {
	      not_closed = false;
	      if (path.front().distance(path.back()) > 1e-7) // if the connexion is not perfect, add a missing edge
		path.addPointAtEnd(path.front());
	    }
	  }
	}
      }
      else if (isFlatEdge(path.back().getFirstVertex(), path.back().getSecondVertex())) {
	assert(path.size() == 1);
	path.addPointAtEnd((*mesh).point(path.back().getFirstVertex()));
      }
      else {
	// TODO: if it's the first edge, check the two sides, and select the farest one
	TriangleID t = (*mesh).findTriangle(path.back().getFirstVertex(), path.back().getSecondVertex()).getId();
	if (path.size() != 1) {
	  TriangleID t2 = (*mesh).findTriangle(path.back(), path[path.size() - 2]).getId();
	  if (t == t2) {
	    try {
	      t = (*mesh).findOtherTriangle(path.back().getFirstVertex(), path.back().getSecondVertex(), t).getId();
	    }
	    catch (Exception e) {
	      if (firstSide) {
		firstSide = false;
		path.reversePath();
		reverse = true;
	      }
	      else
		not_closed = false;
	    }
	  }
	}
	else {
	  // look to the other triangle...
	  try {
	    TriangleID t2 = (*mesh).findOtherTriangle(path.back().getFirstVertex(), path.back().getSecondVertex(), t).getId();
	    const PointOnEdge nextp1 = getNextPointOnLevelSet(path.back(), t, lsvalue);
	    const PointOnEdge nextp2 = getNextPointOnLevelSet(path.back(), t2, lsvalue);
	    if ((nextp1.inSameEdge(path.back())) || ((nextp1.distance(path.back()) < nextp2.distance(path.back())) && (!nextp2.inSameEdge(path.back()))))
	      t = t2;

	  }
	  catch (Exception e) {
	  }
	}

	if (not_closed && !reverse) {
	  const PointOnEdge nextp = getNextPointOnLevelSet(path.back(), t, lsvalue);
	  assert(nextp.inSameEdge(path.back()) ||
		 (t == (*mesh).findTriangle(path.back(), nextp).getId()));
	  path.addPointAtEnd(nextp);
	  if (path.front().inSameEdge(path.back()) && (path.size() > 2)) {
	    not_closed = false;
	    if (path.front().distance(path.back()) > 1e-7) // if the connexion is not perfect, add a missing edge
	      path.addPointAtEnd(path.front());
	  }
	}
      }
    }

    return path;
  }

  template <typename T>
  PointOnEdge MeshMapT<T>::getPossibleNextPointOnLevelSet(const PointOnEdge & point, double lsvalue) const {
    if (mesh == NULL)
      throw Exception("getPossibleNextPointOnLevelSet(2): mesh not defined in MeshMap");
    assert(point.isVertex());
    const Point3D & p = (*mesh).point(point.getClosestVertex());
    VertexID prednb = p.getNeighbours().back();
    for(std::deque<VertexID>::const_iterator nb = p.getNeighbours().begin(); nb != p.getNeighbours().end(); ++nb) {
      assert((*mesh).point(*nb).hasNeighbour(prednb));
      T valuenb = values[*nb];
      T valueprednb = values[prednb];
      if ((valuenb != valueprednb) &&
	  (((valuenb >= lsvalue) && (valueprednb <= lsvalue)) ||
	   ((valuenb <= lsvalue) && (valueprednb >= lsvalue)))) {
	T minvalue = valuenb < valueprednb ? valuenb : valueprednb;
	T maxvalue = valuenb > valueprednb ? valuenb : valueprednb;
	return PointOnEdge(valuenb < valueprednb ? (*mesh).point(*nb) : (*mesh).point(prednb),
			   valuenb < valueprednb ? (*mesh).point(prednb) : (*mesh).point(*nb),
			   (lsvalue - minvalue) / (maxvalue - minvalue));
      }
      prednb = *nb;
    }
    throw Exception("getPossibleNextPointOnLevelSet(2): unable to find next point.");
  }

  template <typename T>
  PointOnEdge MeshMapT<T>::getOtherPossibleNextPointOnLevelSet(const PointOnEdge & point, double lsvalue, const PointOnEdge & pointfirst) const {
    if (mesh == NULL)
      throw Exception("getOtherPossibleNextPointOnLevelSet(3): mesh not defined in MeshMap");
    assert(point.isVertex());
    const Point3D & p = (*mesh).point(point.getClosestVertex());
    VertexID prednb = p.getNeighbours().back();
    for(std::deque<VertexID>::const_iterator nb = p.getNeighbours().begin(); nb != p.getNeighbours().end(); ++nb) {
      assert((*mesh).point(*nb).hasNeighbour(prednb));
      if (!pointfirst.inEdge(std::pair<VertexID, VertexID>(*nb, prednb))) {
	T valuenb = values[*nb];
	T valueprednb = values[prednb];
	if ((valuenb != valueprednb) &&
	    (((valuenb >= lsvalue) && (valueprednb <= lsvalue)) ||
	     (((valuenb <= lsvalue) && (valueprednb >= lsvalue))))) {
	  T minvalue = valuenb < valueprednb ? valuenb : valueprednb;
	  T maxvalue = valuenb > valueprednb ? valuenb : valueprednb;
	  return PointOnEdge(valuenb < valueprednb ? (*mesh).point(*nb) : (*mesh).point(prednb),
			     valuenb < valueprednb ? (*mesh).point(prednb) : (*mesh).point(*nb),
			     (lsvalue - minvalue) / (maxvalue - minvalue));
	}
      }
      prednb = *nb;
    }
    throw Exception("getOtherPossibleNextPointOnLevelSet(2): unable to find next point.");
  }

  template <typename T>
  PointOnEdge MeshMapT<T>::getPointOnEdgeFromValueAndEdge(VertexID p1, VertexID p2, T value) const {
    const T v1 = values[p1];
    const T v2 = values[p2];
    // reorder the values
    if (v1 > v2)
      return getPointOnEdgeFromValueAndEdge(p2, p1, value);
    // throw exception if out of bound
    if (((v1 > value) && v2 > value) ||
	((v2 < value) && v2 < value))
      throw Exception("getPointOnEdgeFromValueAndEdge(3): out of bounds");
    assert(v1 <= v2);
    assert(value <= v2);
    assert(v1 <= value);
    // if the two points have the same value... Choose the middle.
    if (v1 == v2)
      return PointOnEdge((*mesh).point(p1), (*mesh).point(p2), 0.5);
    else // otherwise, adjust the location according to the values
      return PointOnEdge((*mesh).point(p1), (*mesh).point(p2), (value - v1) / (v2 - v1));
  }


  template <typename T>
  PLPath MeshMapT<T>::computeLevelSetFromValueAndPath(const PLPath & path, double value) const {
    return computeLevelSetCC(getPointOnEdgeFromValueAndPath(path, value));
  }

  template <typename T>
  PointOnEdge MeshMapT<T>::getPointOnEdgeFromValueAndPath(const PLPath & path, double value) const {
    if (path.size() == 0)
      throw Exception("getPointOnEdgeFromValueAndPath(2): empty path");
    if (path.size() == 1) {
      if (getValue(path.front()) == value)
	return path.front();
      else
	throw Exception("getPointOnEdgeFromValueAndPath(2): cannot find the required value along the path.");
    }
    PLPath::const_iterator predp = path.begin();
    double predValue = getValue(*predp);
    for(PLPath::const_iterator p = path.begin() + 1; p != path.end(); ++p, ++predp) {
      const double valueP = getValue(*p);
      if (predValue == value)
	return *predp;
      else if (valueP == value)
	return *p;
      else if (((predValue <= value) && (value <= valueP)) ||
	       ((predValue >= value) && (value >= valueP))) {
	if ((*predp).inSameEdge(*p)) {
	  std::pair<VertexID, VertexID> edge = (*predp).getEdge(*p);
	  return PointOnEdge((*mesh).point(edge.first), (*mesh).point(edge.second), (value - values[edge.first]) / (values[edge.second] - values[edge.first]));
	}
	else {
	  const Triangle & t = (*mesh).findTriangle(*predp, *p);
	  for(unsigned int i = 0; i < 3; ++i) {
	    const std::pair<VertexID, VertexID> edge = t.getEdge(i);
	    const double v1 = (*this)[edge.first];
	    const double v2 = (*this)[edge.second];
	    if (((v1 <= value) && (value <= v2)) ||
	       ((v1 >= value) && (value >= v2))) {
	      return PointOnEdge((*mesh).point(edge.first), (*mesh).point(edge.second), (value - v1) / (v2 - v1));
	    }
	  }
	  throw Exception("getPointOnEdgeFromValueAndPath(2): cannot find the required value along the path.");
	}
      }
      predValue = valueP;
    }

    throw Exception("getPointOnEdgeFromValueAndPath(2): cannot find the required value along the path.");
  }


  template <typename T>
  PointOnEdge MeshMapT<T>::getNextPointOnLevelSet(const PointOnEdge & point, TriangleID t, double lsvalue) const {
    if (mesh == NULL)
      throw Exception("getNextPointOnLevelSet(2): mesh not defined in MeshMap");
    const Triangle & triangle = (*mesh).triangle(t);
    VertexID p1 = point.getFirstVertex();
    assert(triangle.hasPoint(p1));
    VertexID p2 = point.getSecondVertex();
    assert(triangle.hasPoint(p2));
    VertexID p3 = triangle.getOtherPoint(p1, p2);

    // reorder p1 and p2: values[p1] <= values[p2]
    if (values[p1] > values[p2]) {
      VertexID ptmp = p1;
      p1 = p2;
      p2 = ptmp;
    }

    if (lsvalue == values[p3]) {
      return (*mesh).point(p3);
    }
    if (lsvalue < values[p3]) {
      // the new point is between p1 and p3
      if (values[p1] < values[p3])
	return PointOnEdge((*mesh).point(p1), (*mesh).point(p3), (lsvalue - values[p1]) / (values[p3] - values[p1]));
      else
	return PointOnEdge((*mesh).point(p3), (*mesh).point(p1), (lsvalue - values[p3]) / (values[p1] - values[p3]));
    }
    else {
      // the new point is between p2 and p3
      if (values[p3] < values[p2])
	return PointOnEdge((*mesh).point(p3), (*mesh).point(p2), (lsvalue - values[p3]) / (values[p2] - values[p3]));
      else
	return PointOnEdge((*mesh).point(p2), (*mesh).point(p3), (lsvalue - values[p2]) / (values[p3] - values[p2]));
    }
  }

  template <typename T>
  std::vector<PLPath> MeshMapT<T>::computeLevelSet(double value) const {
    if (mesh == NULL)
      throw Exception("computeLevelSet(1): mesh not defined in MeshMap");
    std::vector<PLPath> result;

    (*mesh).setTriangleFlag(0);
    (*mesh).setPointFlag(0);
    // for all the triangles
    for(Mesh::triangle_iterator t = (*mesh).triangle_begin(); t != (*mesh).triangle_end(); ++t)
      if ((*t).getFlag() == 0) {
	VertexID inf = (*mesh).getNbPoints();
	VertexID sup = (*mesh).getNbPoints();
	for(unsigned int i = 0; i < 3; ++i) {
	  T v = values[(*t).getVertexId(i)];
	  if (v <= value) inf = (*t).getVertexId(i);
	  if (v >= value) sup = (*t).getVertexId(i);
	}
	// a valid edge is found
	if ((inf != (*mesh).getNbPoints()) && (sup != (*mesh).getNbPoints())) {
	  PointOnEdge firstPoint((*mesh).point(inf), (*mesh).point(sup),
				 values[sup] == values[inf] ? 0. : (value - values[inf]) / (values[sup] - values[inf]));
	  if (!firstPoint.isVertex() || ((*mesh).point(firstPoint.getClosestVertex()).getFlag() == 0)) {
	    assert((*mesh).getBox().contains(firstPoint));
	    PLPath path = computeLevelSetCC(firstPoint);
	    (*mesh).setTriangleAndPointFlag(path, 1);
	    result.push_back(path);
	  }
	}
      }

    return result;
  }

  template <typename T>
  bool MeshMapT<T>::isSaddlePoint(VertexID p) const {
    return getNeighbourArity((*mesh).point(p)) > 1;
  }

  template <typename T>
  std::vector<std::pair<VertexID, unsigned int> > MeshMapT<T>::getSaddlePoints() const {
    if (mesh == NULL)
      throw Exception("getSaddlePoints(): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("getSaddlePoints(): values not defined.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("getSaddlePoints(): mesh has been modified (bad number of values).");

    std::vector<std::pair<VertexID, unsigned int> > result;

    T * v = values;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p, ++v)
      if ((*p).getNbNeighbours() != 0) {
	unsigned int arity = getNeighbourArity(*p);

	if (arity > 1)
	  result.push_back(std::pair<VertexID, unsigned int>((*p).getId(), arity));
      }

    return result;
  }

  template <typename T>
  unsigned int MeshMapT<T>::getNeighbourArity(const Point3D & p, T * curValue) const {
    unsigned int arity = 0;
    if (curValue == NULL)
      curValue = values + p.getId();
    VertexID prevnb = p.getNeighbours().back();
    for(std::deque<VertexID>::const_iterator nb = p.getNeighbours().begin(); nb != p.getNeighbours().end(); prevnb = *nb, ++nb)
      if ((values[*nb] > *curValue) && (values[prevnb] < *curValue))
	++arity;
    return arity;
  }

  template <typename T>
  std::vector<VertexID> MeshMapT<T>::getLocalExtrema() const {
    if (mesh == NULL)
      throw Exception("getLocalExtrema(): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("getLocalExtrema(): values not defined.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("getLocalExtrema(): mesh has been modified (bad number of values).");

    std::vector<VertexID> result;

    T * v = values;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p, ++v)
      if ((*p).getNbNeighbours() != 0) {
	bool isExtrema = true;
	bool min = values[(*p).getNeighbours().front()] > *v;

	for(std::deque<VertexID>::const_iterator nb = (*p).getNeighbours().begin(); nb != (*p).getNeighbours().end(); ++nb)
	  if ((min && (values[*nb] <= *v)) ||
	      ((!min) && (values[*nb] >= *v))) {
		isExtrema = false;
		break;
	      }
	if (isExtrema)
	  result.push_back((*p).getId());
      }

    return result;
  }

  template <typename T>
  VertexID MeshMapT<T>::getNbFlatEdges() const {
    if (mesh == NULL)
      throw Exception("getNbFlatEdges(): mesh not defined in MeshMap");
    if (values == NULL)
      throw Exception("getNbFlatEdges(): values not defined.");
    if ((*mesh).getNbPoints() != nbValues)
      throw Exception("getNbFlatEdges(): mesh has been modified (bad number of values).");

    VertexID result = 0;

    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p)
      for(std::deque<VertexID>::const_iterator nb = (*p).getNeighbours().begin(); nb != (*p).getNeighbours().end(); ++nb)
	if ((*nb > (*p).getId()) && (values[*nb] == values[(*p).getId()]))
	  ++result;

    return result;
  }


  template <typename T>
  T MeshMapT<T>::getValue(const PointOnEdge & point) const {
    const T & v = values[point.getFirstVertex()];
    return v + point.getLocation() * (values[point.getSecondVertex()] - v);
  }

  template <typename T>
  MeshMapT<T> MeshMapT<T>::distanceMap(T value) const {
    MeshMapT<T> result(*this);
    assert(result.getNbValues() == nbValues);
    T * v = result.values;
    for(VertexID i = 0; i < nbValues; ++i, ++v)
      *v = fabs(*v - value);
    return result;
  }

  template <typename T>
  T MeshMapT<T>::getMinimumDifferenceOnEdge() const {
    T result = std::numeric_limits<T>::max();

    T * v = values;
    for(Mesh::const_point_iterator p = (*mesh).point_begin(); p != (*mesh).point_end(); ++p, ++v)
      for(std::deque<VertexID>::const_iterator nb = (*p).getNeighbours().begin(); nb != (*p).getNeighbours().end(); ++nb) {
	const T diff = fabs(*v - values[*nb]);
	if ((diff != 0) && (diff < result))
	  result = diff;
      }
    return result;
  }

  template <typename T>
  double MeshMapT<T>::getArea(T minValue, T maxValue) const {
    return getArea(minValue, true) + getArea(maxValue, false) - (*mesh).getArea();
  }

  template <typename T>
  double MeshMapT<T>::getArea(T value, bool sup) const {
    double result = 0.;
    double direction = sup ? -1 : 1;

    for(Mesh::const_triangle_iterator t = (*mesh).triangle_begin(); t != (*mesh).triangle_end(); ++t) {
      const bool p1Inside = (value - values[(*t).getP1()]) * direction > 0;
      const bool p2Inside = (value - values[(*t).getP2()]) * direction > 0;
      const bool p3Inside = (value - values[(*t).getP3()]) * direction > 0;
      const unsigned int nbInside = (unsigned int)p1Inside + (unsigned int)p2Inside + (unsigned int)p3Inside;
      if (nbInside == 3)
	result += (*mesh).getArea((*t).getId());
      else if (nbInside == 2) {
	Coord3D c1, c2, c3;
	if (!p1Inside)
	  c1 = (*mesh).point((*t).getP1());
	else
	  if (!p2Inside)
	    c1 = getPointOnEdgeFromValueAndEdge((*t).getP1(), (*t).getP2(), value);
	  else // !p3Inside
	    c1 = getPointOnEdgeFromValueAndEdge((*t).getP1(), (*t).getP3(), value);
	if (!p2Inside)
	  c2 = (*mesh).point((*t).getP2());
	else
	  if (!p1Inside)
	    c2 = getPointOnEdgeFromValueAndEdge((*t).getP2(), (*t).getP1(), value);
	  else // !p3Inside
	    c2 = getPointOnEdgeFromValueAndEdge((*t).getP2(), (*t).getP3(), value);
	if (!p3Inside)
	  c3 = (*mesh).point((*t).getP3());
	else
	  if (!p1Inside)
	    c3 = getPointOnEdgeFromValueAndEdge((*t).getP3(), (*t).getP1(), value);
	  else // !p2Inside
	    c3 = getPointOnEdgeFromValueAndEdge((*t).getP3(), (*t).getP2(), value);

	const Triangle3D tt(c1, c2, c3);
	result += (*mesh).getArea((*t).getId()) - tt.area();
      }
      else if (nbInside == 1) {
	Coord3D c1, c2, c3;
	if (p1Inside)
	  c1 = (*mesh).point((*t).getP1());
	else
	  if (p2Inside)
	    c1 = getPointOnEdgeFromValueAndEdge((*t).getP1(), (*t).getP2(), value);
	  else // p3Inside
	    c1 = getPointOnEdgeFromValueAndEdge((*t).getP1(), (*t).getP3(), value);
	if (p2Inside)
	  c2 = (*mesh).point((*t).getP2());
	else
	  if (p1Inside)
	    c2 = getPointOnEdgeFromValueAndEdge((*t).getP2(), (*t).getP1(), value);
	  else // p3Inside
	    c2 = getPointOnEdgeFromValueAndEdge((*t).getP2(), (*t).getP3(), value);
	if (p3Inside)
	  c3 = (*mesh).point((*t).getP3());
	else
	  if (p1Inside)
	    c3 = getPointOnEdgeFromValueAndEdge((*t).getP3(), (*t).getP1(), value);
	  else // p2Inside
	    c3 = getPointOnEdgeFromValueAndEdge((*t).getP3(), (*t).getP2(), value);

	const Triangle3D tt(c1, c2, c3);
	result += tt.area();
      }
    }

    return result;
  }
}

#endif
